"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio

import pytest

from inmanta import data
from inmanta.server import config as opt
from inmanta.server.services import databaseservice
from utils import retry_limited


@pytest.mark.asyncio
async def test_agent_process_cleanup(server, environment, agent_factory):
    opt.agent_processes_to_keep.set("1")
    a1 = await agent_factory(environment, hostname="host", agent_map=[], agent_names=["agent1"])
    a2 = await agent_factory(environment, hostname="host", agent_map=[], agent_names=["agent1"])
    await asyncio.gather(*[a1.stop(), a2.stop()])

    async def _wait_until_expire_is_finished():
        result = await data.AgentProcess.get_list()
        return len([r for r in result if r.expired is not None]) == 2

    await retry_limited(_wait_until_expire_is_finished, timeout=10)
    # Execute cleanup
    database_slice = server.get_slice(databaseservice.SLICE_DATABASE)
    await database_slice._purge_agent_processes()
    # Assert cleanup
    result = await data.AgentProcess.get_list()
    assert len(result) == 1



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import pytest

import inmanta.compiler as compiler
from inmanta.ast import CompilerException


def test_max_iterations(snippetcompiler, monkeypatch):
    monkeypatch.setenv("INMANTA_MAX_ITERATIONS", "1")

    with pytest.raises(CompilerException) as e:
        snippetcompiler.setup_for_snippet(
            """
    import std

    entity Hg:
    end

    Hg.hosts [0:] -- std::Host

    implement Hg using std::none

    hg = Hg()

    for i in [1,2,3]:
     hg.hosts = std::Host(name="Test{{i}}", os=std::unix)
    end


    for i in hg.hosts:
        std::ConfigFile(host=i, path="/fx", content="")
    end
    """
        )

        (types, _) = compiler.do_compile()

    assert "Could not complete model, max_iterations 1 reached." in str(e.value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Inmanta documentation build configuration file, created by
# sphinx-quickstart on Wed Aug 21 10:14:55 2013.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys, os, pkg_resources, datetime

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc', 'sphinx.ext.todo', 'sphinx.ext.coverage', 'sphinx.ext.ifconfig', 'sphinx.ext.viewcode',
    'sphinxarg.ext', 'sphinxcontrib.inmanta.config', 'sphinxcontrib.inmanta.dsl', 'sphinx_tabs.tabs',
    'sphinxcontrib.inmanta.environmentsettings', 'sphinx_click.ext'
]


try:
    # noinspection PyUnresolvedReferences
    # "tags" are injected while the file is being read
    if tags.has("include_redoc"):
        extensions.append('sphinxcontrib.redoc')
except NameError as e:
    # Openapi definition with Redoc won't be included
    pass

redoc_uri = 'https://cdn.jsdelivr.net/npm/redoc/bundles/redoc.standalone.js'
redoc = [
    {
        'name': 'Inmanta REST API',
        'page': 'reference/openapi',
        'spec': 'reference/openapi.json',
        'opts': {
            'hide-hostname': True,
            'path-in-middle-panel': True,
        }
    },
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Inmanta'
copyright = f'{datetime.datetime.now().year} Inmanta NV'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
if "INMANTA_DONT_DISCOVER_VERSION" in os.environ:
    # Used to:
    # 1) Decouple the inmanta-core package from the inmanta
    #    package when running the tests.
    # 2) Build the ISO documentation, which is built on top
    #    of the inmanta-core documentation. During the ISO docs build,
    #    this value will be overwritten with the ISO product version.
    version = "1.0.0"
else:
    version = pkg_resources.get_distribution("inmanta").version
# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# The documentation build tool overrides this when extensions are included in the documentation build.
exclude_patterns = ['extensions.rst']

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

html_theme_options = {
    'logo_only': True,
    'display_version': True,
}
import sphinx_rtd_theme

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/inmanta-logo.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
   '**': ['globaltoc.html', 'sourcelink.html', 'searchbox.html']
}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'InmantaDoc'


# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
# 'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
# 'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
# 'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('index', 'inmanta.tex', 'Inmanta Documentation',
   'Inmanta NV', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'inmanta', 'Inmanta Documentation',
     ['Inmanta NV'], 1)
]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', 'Inmanta', 'Inmanta Documentation',
   'Inmanta NV', 'Inmanta', 'One line description of project.',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'

# Ingnore link check of openapi.html because it's used in a toctree.
# A trick was required to include a non-sphinx document in a toctee.
linkcheck_ignore = [r'http(s)?://localhost:\d+/', r'http://127.0.0.1:\d+', r'openapi.html', r'https://twitter.com/inmanta_com']

# Do not print the warning that tabs only work in html
# https://github.com/djungelorm/sphinx-tabs/issues/39
sphinx_tabs_nowarn = True



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import os
from typing import AsyncIterator

import pytest
from asyncpg import Connection

from db.common import PGRestore
from inmanta.server.bootloader import InmantaBootloader


@pytest.fixture
async def migrate_v5_to_v6(
    hard_clean_db, hard_clean_db_post, postgresql_client: Connection, async_finalizer, server_config
) -> AsyncIterator[None]:
    # Get old tables
    with open(os.path.join(os.path.dirname(__file__), "dumps/v5.sql"), "r") as fh:
        await PGRestore(fh.readlines(), postgresql_client).run()

    ibl = InmantaBootloader()

    await ibl.start()
    # When the bootloader is started, it also executes the migration to v6
    yield
    await ibl.stop()


@pytest.mark.asyncio
async def test_add_on_delete_cascade_constraint(migrate_v5_to_v6, postgresql_client: Connection) -> None:
    """
    Verify that the ON DELETE CASCADE constraint is set correctly on the substitute_compile_id column
    of the compile table.
    """
    # Assert values in substitute_compile_id column are correct
    compiles = await postgresql_client.fetch("SELECT substitute_compile_id FROM public.compile")
    assert all([c["substitute_compile_id"] is None for c in compiles])

    # Assert that ON DELETE CASCADE is set the foreign key constraint compile_substitute_compile_id_fkey
    constraints = await postgresql_client.fetch(
        """
            SELECT pg_catalog.pg_get_constraintdef(r.oid, true) as condef
            FROM pg_catalog.pg_constraint r
            WHERE conname='compile_substitute_compile_id_fkey'
        """
    )
    assert len(constraints) == 1
    assert "ON DELETE CASCADE" in constraints[0]["condef"]



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import re

import pytest

import inmanta.compiler as compiler
from inmanta.ast import DuplicateException


def test_2386_duplicate_attribute_error_message(snippetcompiler) -> None:
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    string test
    bool test
end
        """
    )
    dir: str = snippetcompiler.project_dir
    with pytest.raises(
        DuplicateException,
        match=re.escape(f"attribute already exists (original at ({dir}/main.cf:3:12)) (duplicate at ({dir}/main.cf:4:10))"),
    ):
        compiler.do_compile()



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import os
import uuid
from typing import AsyncIterator

import pytest
from asyncpg import Connection, exceptions

from db.common import PGRestore
from inmanta import data
from inmanta.server.bootloader import InmantaBootloader
from utils import retry_limited


@pytest.fixture
async def migrate_v7_to_v17(
    hard_clean_db, hard_clean_db_post, postgresql_client: Connection, server_config
) -> AsyncIterator[None]:
    """
    Performs a v7 database restore and migrates to v17.
    """
    # Get old tables
    with open(os.path.join(os.path.dirname(__file__), "dumps/v7.sql"), "r") as fh:
        await PGRestore(fh.readlines(), postgresql_client).run()

    ibl = InmantaBootloader()

    await ibl.start()
    # When the bootloader is started, it also executes the migration to v18
    yield
    await ibl.stop()


@pytest.mark.asyncio(timeout=20)
async def test_foreign_key_agent_to_agentinstance(migrate_v7_to_v17: None, postgresql_client: Connection) -> None:
    """
    Deleting an entry in the agentInstance table should not be allowed when it's references from the agent stable.
    """
    env_id = "a21a4bdc-68f0-4c4e-a4f2-a6b42880661a"

    async def is_internal_agent_up() -> bool:
        result = await data.Agent.get_one(environment=env_id, name="internal")
        return result.id_primary is not None

    async def get_id_primary_internal_agent() -> uuid.UUID:
        # Return the id of the AgentInstance that is the primary for the internal agent
        result = await data.Agent.get_one(environment=env_id, name="internal")
        assert result.id_primary is not None
        return result.id_primary

    await retry_limited(is_internal_agent_up, timeout=10)
    id_primary_internal_agent = await get_id_primary_internal_agent()

    # Delete primary agent instance for internal agent should raise an exception
    with pytest.raises(
        exceptions.ForeignKeyViolationError, match=r'violates foreign key constraint "agent_id_primary_fkey" on table "agent"'
    ):
        await postgresql_client.execute(
            f"""
            DELETE FROM public.agentinstance
            WHERE id='{id_primary_internal_agent}';
            """
        )
